﻿using ArcGIS.Core.Data;
using ArcGIS.Core.Data.Realtime;
using ArcGIS.Core.Hosting;
using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace Realtime_SampleCode
{
    class Program
    {
        [STAThread]
        static void Main(string[] args)
        {
            //Call Host.Initialize before constructing any objects from ArcGIS.Core
            Host.Initialize();
            ConnectToStreamServiceAsync();
            Console.Read();
            Console.Read();
        }

        static async Task ConnectToStreamServiceAsync()
        {
            string url = "https://loanr0940837.esri.com:6443/arcgis/rest/services/WaterGauge_NullGeo/StreamServer";                                               //"https://loanr0940837.esri.com/arcgis/rest/services/Florence-Polygon-Out/StreamServer/"
            WriteToConsole($"Connecting to {url}...", ConsoleColor.Yellow);
            var realtimeServiceConProp = new RealtimeServiceConnectionProperties(
                                new Uri(url),
                                RealtimeDatastoreType.StreamService
                            );

            using (var rt_datastore = new RealtimeDatastore(realtimeServiceConProp))
            { 
                var name = rt_datastore.GetTableNames().First();
                using (var realtimeFeatureClass = rt_datastore.OpenTable(name) as RealtimeFeatureClass)
                {
                    realtimeFeatureClass.StartStreaming();
                    WriteToConsole("Streaming started!", ConsoleColor.DarkYellow);
                    
                    realtimeFeatureClass.SetFilterWhereClause("WaterLevel > 10");
                    RealtimeCursor realtimeCursor = realtimeFeatureClass.Subscribe(null, true);

                    while (await realtimeCursor.WaitForRowsAsync())
                    {
                        while (realtimeCursor.MoveNext())
                        {
                            switch (realtimeCursor.Current.GetRowSource())
                            {
                                case RealtimeRowSource.EventInsert:
                                    RealtimeFeature realtimeFeature = realtimeCursor.Current as RealtimeFeature;
                                    //Console.WriteLine($"Gauge ID: {realtimeFeature["GaugeLID"]} \tWater Level: {realtimeFeature["WaterLevel"]}");
                                    await ShowFloodInfoAsync(realtimeFeature);
                                    continue;
                                default:
                                    continue;
                            }
                        }
                    }
                }
            }
        }


        static async Task ShowFloodInfoAsync(RealtimeFeature realtimeFeature)
        {
            await ArcGIS.Core.Hosting.Threading.Tasks.QueuedWorker.Run(() => 
            {
                #region extract stream feature info
                Geodatabase fgdb = new Geodatabase(new FileGeodatabaseConnectionPath(new Uri(_fgdbPath)));
                string gaugeId = realtimeFeature["GaugeLID"].ToString().Trim();

                FeatureClass streamGaugeFC = fgdb.OpenDataset<FeatureClass>("StreamGauges");
                var queryFilter = new QueryFilter()
                {
                    WhereClause = $"GaugeLID = '{gaugeId}'"
                };
                RowCursor streamGaugeRowCursor = streamGaugeFC.Search(queryFilter, true);
                Feature streamGaugeFeature = null;
                if (streamGaugeRowCursor.MoveNext())
                    streamGaugeFeature = streamGaugeRowCursor.Current as Feature;

                if (streamGaugeFeature != null)
                {
                    string floodStatus = "Unknown";
                    double waterLevel = double.Parse(realtimeFeature["WaterLevel"].ToString());
                    double actionWL = (streamGaugeFeature["Action"] == null) ? 0 : double.Parse(streamGaugeFeature["Action"].ToString());
                    double floodWL = (streamGaugeFeature["Flood"] == null) ? actionWL : double.Parse(streamGaugeFeature["Flood"].ToString());
                    double moderateWL = (streamGaugeFeature["Moderate"] == null) ? floodWL : double.Parse(streamGaugeFeature["Moderate"].ToString());
                    double majorWL = (streamGaugeFeature["Major"] == null) ? moderateWL : double.Parse(streamGaugeFeature["Major"].ToString());
                    if (waterLevel > majorWL)
                        floodStatus = "Major Flood";
                    else if (waterLevel > moderateWL)
                        floodStatus = "Moderate Flood";
                    else if (waterLevel > floodWL)
                        floodStatus = "Flood";
                    else
                        floodStatus = "Warning";


                    #endregion

                    var guage_location = streamGaugeFeature.GetShape();
                    FeatureClass countyFC = fgdb.OpenDataset<FeatureClass>("Counties");
                    SpatialQueryFilter sf = new SpatialQueryFilter()
                    {
                        FilterGeometry = guage_location,
                        SpatialRelationship = SpatialRelationship.Within
                    };
                    RowCursor countyRowCursor = countyFC.Search(sf, true);
                    if (countyRowCursor.MoveNext())
                    {
                        #region Display info on Console
                        Row countyRow = countyRowCursor.Current;
                        Console.ResetColor();
                        Console.Write($"Gauge ID: {gaugeId.PadRight(8)} Water Level: {realtimeFeature["WaterLevel"].ToString().PadRight(5)}");
                        Console.ForegroundColor = ConsoleColor.Blue;
                        Console.Write($"Flood Status: {floodStatus.PadRight(16)} County: {countyRow["Name"].ToString().PadRight(18)} State: {countyRow["State_Name"].ToString().PadRight(16)} Population: {countyRow["POP2010"]}\n");
                        Console.ResetColor();
                    }
                    else
                    {
                        Console.WriteLine($"Gauge ID: {gaugeId.PadRight(8)} Water Level: {realtimeFeature["WaterLevel"].ToString().PadRight(5)} Flood Status: {floodStatus.PadRight(16)} {{County info not found}}");
                        #endregion
                    }
                }
                else
                {
                    Console.WriteLine($"Gauge ID: {gaugeId.PadRight(8)} Water Level: {realtimeFeature["WaterLevel"].ToString().PadRight(5)} Flood Status: Unknown");
                }
            });
        }

        #region variables and helper functions
        static string _fgdbPath = @"C:\Projects\DevSummit\2020\Stream Layer\Realtime_SampleCode\Data.gdb";
        static void WriteToConsole(string msg, ConsoleColor color)
        {
            Console.ForegroundColor = color;
            Console.WriteLine(msg);
            Console.ResetColor();
        }
        #endregion
    }
}
